package {
	import net.flashpunk.*;
	import net.flashpunk.utils.*;
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.PixelSnapping;
	import flash.geom.ColorTransform;
	import flash.geom.Point;

	/** Transition between worlds */
	public class Transition extends World {

		public static var time:Number = 0.6;	// Default transition time

		private var worldA: World;
		private var worldB: World;
		private var value: Number;
		private var step: Number;
		private var updateA:Boolean;
		private var updateB:Boolean;

		private var colour: uint=0;
		private var direction: Point;
		private var changeCallback:Function;

		private var bufferA: Bitmap;
		private var bufferB: Bitmap;
		private var cTransform: ColorTransform;

		private var mode:int;
		private static const FADEOUT:	int=1;
		private static const CROSSFADE:int=2;
		private static const SLIDE:	int=3;


		/** Start a fadout transition. Fades to solid colour then to new world.
		 * @param next		The world to transition to
		 * @param colour	The colour to fade to
		 * @param time		Duration of the transition in seconds
		 * @param change	Function to call when the state is solid colour
		 * @param updateA	Update current world while it is fading
		 * @param updateB	Update next world while it is fading
		 */
		public static function fadeOut(next:World, colour:uint=0, time:Number=0, change:Function=null, updateA:Boolean=false, updateB:Boolean=false): void {
			var t:Transition = new Transition(FP.world, next);
			t.colour = colour;
			t.duration = time>0? time: Transition.time;
			t.updateA = updateA;
			t.updateB = updateB;
			t.changeCallback = change;
			t.mode = FADEOUT;
			FP.world = t;
		}
		/** Start a crossfade transition
		 * @param next		The world to transition to
		 * @param time		Duration of the transition in seconds
		 * @param updateA	Update current world while it is fading
		 * @param updateB	Update next world while it is fading
		 */ 
		public static function crossFade(next:World, time:Number=0.6, updateA:Boolean=false, updateB:Boolean=false): void {
			var t:Transition = new Transition(FP.world, next);
			t.cTransform = new ColorTransform();
			t.duration = time>0? time: Transition.time;
			t.updateA = updateA;
			t.updateB = updateB;
			t.mode = CROSSFADE;
			FP.world = t;
		}
		/** Start an upwards sliding transition
		 * @param next		The world to transition to
		 * @param direction	Direction to slide the worlds [ UP, DOWN, LEFT, RIGHT ]
		 * @param time		Duration of the transition in seconds
		 * @param updateA	Update current world while it is fading
		 * @param updateB	Update next world while it is fading
		 */ 
		public static function slideUp(next:World, time:Number=0.6, updateA:Boolean=false, updateB:Boolean=false): void {
			var t:Transition = new Transition(FP.world, next);
			t.duration = time>0? time: Transition.time;
			t.updateA = updateA;
			t.updateB = updateB;
			t.mode = SLIDE;
			t.direction = new Point(0, -1);
			FP.world = t;
		}
		/** Start a downwards sliding transition
		 * @param next		The world to transition to
		 * @param direction	Direction to slide the worlds [ UP, DOWN, LEFT, RIGHT ]
		 * @param time		Duration of the transition in seconds
		 * @param updateA	Update current world while it is fading
		 * @param updateB	Update next world while it is fading
		 */ 
		public static function slideDown(next:World, time:Number=0.6, updateA:Boolean=false, updateB:Boolean=false): void {
			var t:Transition = new Transition(FP.world, next);
			t.duration = time>0? time: Transition.time;
			t.updateA = updateA;
			t.updateB = updateB;
			t.mode = SLIDE;
			t.direction = new Point(0, 1);
			FP.world = t;
		}
	
		/** Start a left sliding transition
		 * @param next		The world to transition to
		 * @param direction	Direction to slide the worlds [ UP, DOWN, LEFT, RIGHT ]
		 * @param time		Duration of the transition in seconds
		 * @param updateA	Update current world while it is fading
		 * @param updateB	Update next world while it is fading
		 */ 
		public static function slideLeft(next:World, time:Number=0.6, updateA:Boolean=false, updateB:Boolean=false): void {
			var t:Transition = new Transition(FP.world, next);
			t.duration = time>0? time: Transition.time;
			t.updateA = updateA;
			t.updateB = updateB;
			t.mode = SLIDE;
			t.direction = new Point(-1, 0);
			FP.world = t;
		}
		/** Start a right sliding transition
		 * @param next		The world to transition to
		 * @param direction	Direction to slide the worlds [ UP, DOWN, LEFT, RIGHT ]
		 * @param time		Duration of the transition in seconds
		 * @param updateA	Update current world while it is fading
		 * @param updateB	Update next world while it is fading
		 */ 
		public static function slideRight(next:World, time:Number=0.6, updateA:Boolean=false, updateB:Boolean=false): void {
			var t:Transition = new Transition(FP.world, next);
			t.duration = time>0? time: Transition.time;
			t.updateA = updateA;
			t.updateB = updateB;
			t.mode = SLIDE;
			t.direction = new Point(1, 0);
			FP.world = t;
		}

		/** Constructor. Use static methods to invoke transitions */
		public function Transition(a:World, b:World) {
			worldA = a;
			worldB = b;
			b.updateLists();
			value = 0;
		}
		public override function begin(): void {
			bufferA = createTarget();
			bufferB = createTarget();
		}

		// Set transition duration in seconds
		private function set duration(t:Number): void {
			if(t==0) step=1;
			else step = 1/60/t;
		}

		// Create a render target
		private function createTarget(): Bitmap {
			return new Bitmap( new BitmapData(FP.width, FP.height, false, 0), PixelSnapping.NEVER);
		}

		// Update transition
		public override function update(): void {
			value += step;

			// Update worlds
			if(mode==FADEOUT) {
				if(value>=0.5 && value-step<0.5) {
					if(changeCallback!=null) changeCallback();
				}
				if(updateA && value<0.5)  updateWorld(worldA);
				if(updateB && value>=0.5) updateWorld(worldB);
			} else {
				if(updateA) updateWorld(worldA);
				if(updateB) updateWorld(worldB);
			}

			//Transition Complete
			if(value >= 1) {
				value = 1;
				FP.world = worldB;
			}
		}

		// Update a world
		private function updateWorld(world:World): void {
			world.update();
			world.updateLists();
		}
		
		// render a world to a buffer
		private function renderWorld(world:World, target:*): void {
			var buffer:BitmapData = FP.buffer;
			if(target is Bitmap) FP.buffer = target.bitmapData;
			else if(target is BitmapData) FP.buffer = target;
			FP.camera = world.camera;
			Draw.resetTarget();
			world.render();
			FP.buffer = buffer;
			FP.camera = camera;
		}

		// Render stuff
		public override function render(): void {
			switch(mode) {
			case FADEOUT:
				var alpha:Number = 1 - Math.abs(value-0.5) * 2;
				renderWorld(value<0.5? worldA: worldB, FP.buffer);
				Draw.resetTarget();
				Draw.rect(0,0,FP.width, FP.height, colour, alpha);
				break;

			case CROSSFADE:
				worldA.render();
				renderWorld(worldA, FP.buffer);
				renderWorld(worldB, bufferB);
				cTransform.alphaMultiplier = value;
				FP.buffer.copyPixels(bufferA.bitmapData, FP.bounds, FP.zero);
				FP.buffer.draw(bufferB.bitmapData, null, cTransform);
				break;

			case SLIDE:
				renderWorld(worldA, bufferA);
				renderWorld(worldB, bufferB);

				FP.point.x = direction.x*value * FP.width;
				FP.point.y = direction.y*value * FP.height;
				FP.buffer.copyPixels(bufferA.bitmapData, FP.bounds, FP.point);

				FP.point.x += -direction.x * FP.width;
				FP.point.y += -direction.y * FP.height;
				FP.buffer.copyPixels(bufferB.bitmapData, FP.bounds, FP.point);
				break;
			}
		}
	}
}

